package com.pact.com.pact.test;

import static org.junit.jupiter.api.Assertions.*;

import com.pact.Node;
import com.pact.Nodes;
import org.junit.jupiter.api.Test;

import java.util.HashSet;

class NodeTest {

    @Test
    public void equalsTest1(){
        Node node1 = new Node(null, "(ABC)");
        Node node2 = new Node(null, "(AB)");
        assertFalse(node1.equals(node2));
    }

    @Test
    public void equalsTest2(){
        Node node1 = new Node(null, "(ABC)");
        Node node2 = new Node(null, "(AB)");
        assertFalse(node2.equals(node1));
    }

    @Test
    public void equalsTest3(){
        Node node1 = new Node(null, "((BC)A)");
        Node node2 = new Node(null, "(A(CB))");
        assertTrue(node1.equals(node2));
    }

    @Test
    public void equalsTest4(){
        Node node1 = new Node(null, "(A(BC))");
        Node node2 = new Node(null, "(A(BCD))");
        assertFalse(node1.equals(node2));
    }

    @Test
    public void searchTest1() {
        Node node1 = new Node(null, "(A(B(C(DA))))");
        Node node2 = new Node(null, "(A(B(CD)))");
        HashSet<Nodes> result = new HashSet<>();
        node2.deepSearch(node1, result);
        System.out.println(result);
    }

    void reducedCombineTest(String s1, String s2, String sResult){
        Node node1 = new Node(null, s1);
        Node node2 = new Node(null, s2);
        HashSet<Node> results = node1.reducedCombine(node2);
        for(Node result : results)
            System.out.println("" + node1 + " + " + node2 + " => " + result);
        assertTrue(results.size() == 1);
        Node result = results.iterator().next();
        assertEquals(new Node(null, sResult), result);
    }

    void reducedCombineTest(String s1, String s2, String[] sResults){
        Node node1 = new Node(null, s1);
        Node node2 = new Node(null, s2);
        HashSet<Node> results = node1.reducedCombine(node2);
        for(Node result : results)
            System.out.println("" + node1 + " + " + node2 + " => " + result);
        assertTrue(results.size() == sResults.length);
        for (String sResult : sResults) {
            Node testNode = new Node(null, sResult);
            boolean found = false;
            for(Node node : results) {
                if(testNode.equals(node)) {
                    found = true;
                    break;
                }
            }
            assertTrue(found);
        }
    }

    @Test
    public void reducedCombineTest1() {
        reducedCombineTest("(DB)", "(AC)","((DB)(AC))");
    }

    @Test
    public void reducedCombineTest2() {
        reducedCombineTest("(A(BDC))", "((CDB)A)", "(A(BDC))");
    }


    @Test
    public void reducedCombineTest3() {
        reducedCombineTest("(A(CD))", "(A(B(CD)))", "(A(B(CD)))");
    }

    @Test
    public void reducedCombineTest4() {
        reducedCombineTest("(A((BE)(CD)))", "(A(B(CD)))", "(A((BE)(CD)))");
    }

    @Test
    public void reducedCombineTest5() {
        reducedCombineTest("(A(A(B(CD))))", "(A((BE)(C(DA))))", "((((BE)(C(DA)))A)A)");
    }

    @Test
    public void reducedCombineTest6() {
        reducedCombineTest("(A(B(CD)))", "((A(B(CD)))(A(B(CD))))", "(A(B(CD)))");
    }

    @Test
    public void reducedCombineTest7() {
        reducedCombineTest("(A(CD))", "(A(B(C(DA))))", "(A(B(C(DA))))");
    }

    @Test
    public void reducedCombineTest8() {
        reducedCombineTest("(A(B(CD)))", "(A(B(C(DA))))", "(A(B(C(DA))))");
    }

    @Test
    public void reducedCombineTest9() {
        reducedCombineTest("A", "B", "(AB)");
    }

    @Test
    public void reducedCombineTest10() {
        reducedCombineTest("(A(B(D((CE)(CD)))))", "(F(C(DB)((D(BE))(CD))))", "(((((CD)((ECB)D))C(DB))BF)A)");
    }

    @Test
    public void reducedCombineTest10b() {
        reducedCombineTest("(A(B(D((CE)(CD)))))", "(F(C(DB)((D(BE))(CD))))", "(((((CD)((ECB)D))C(DB))BF)A)");
    }

    @Test
    public void reducedCombineTest11() {
        reducedCombineTest("((AB)(CD))", "(E(F(GA)))", "(((AB)(CD))(E(F(GA))))");
    }

    @Test
    public void reducedCombineTest12() {
        reducedCombineTest("(A(B(CD)))", "(D(C(BA)))", "((A(B(CD)))(D(C(BA))))");
    }

    @Test
    public void reducedCombineTest13() {
        reducedCombineTest("(A(CD))", "(A(BE))", "(A(CD)(BE))");
    }

    @Test
    public void reducedCombineTest14() {
        reducedCombineTest("(A(CD)(BE))", "(A((CD)(BE)))", "(A((CD)(BE)))");
    }

    @Test
    public void reducedCombineTest15() {
        String results[] = {"((AC)(AB))", "(((AC)B)A)"};
        reducedCombineTest("((AC)B)", "(A(AB))", results);
    }

    @Test
    public void reducedCombineTest16() {
        String results[] = {"(((BA)C)A)", "(((CA)B)A)"};
        reducedCombineTest("((AC)BA)", "(A(CB))", results);
    }

    @Test
    public void reducedCombineTest17() {
        reducedCombineTest("(A(BC))", "(DB)", "((A(BC))(DB))");
    }

    @Test
    public void reducedCombineTest18() {
        reducedCombineTest("(((TH)(IS))(R(E(AL))))", "(((TH)(IS))(T(RE)))",
                           "(((TH)(IS))(((E(AL))R)T))");
    }


}